'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _toConsumableArray2 = require('babel-runtime/helpers/toConsumableArray');

var _toConsumableArray3 = _interopRequireDefault(_toConsumableArray2);

var _from = require('babel-runtime/core-js/array/from');

var _from2 = _interopRequireDefault(_from);

var _clz = require('babel-runtime/core-js/math/clz32');

var _clz2 = _interopRequireDefault(_clz);

exports.isArrayEqualWith = isArrayEqualWith;
exports.getRandomInt = getRandomInt;
exports.getRandomBigInt = getRandomBigInt;
exports.getRandomSecrets = getRandomSecrets;
exports.getRandomPoints = getRandomPoints;
exports.getSecretHash = getSecretHash;
exports.shuffleArray = shuffleArray;
exports.pointToJSON = pointToJSON;
exports.sortPoints = sortPoints;

var _bn = require('bn.js');

var _bn2 = _interopRequireDefault(_bn);

var _crypto = require('crypto');

var _crypto2 = _interopRequireDefault(_crypto);

var _config = require('./config');

var _config2 = _interopRequireDefault(_config);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Utility functions.
 * @namespace Utils
 */

/**
 * Determines whether the given arrays are identical.
 * @memberof Utils
 * @param {T[]} array1 The first array to compare.
 * @param {T[]} array2 The second array to compare.
 * @param {Function} [compareFunction] Specifies a function that defines the
 * method of individual value comparison.
 * @returns {boolean}
 */
function isArrayEqualWith(array1, array2, compareFunction) {
  /* eslint-disable no-param-reassign */
  compareFunction = compareFunction || function (a, b) {
    return a === b;
  };
  /* eslint-enable no-param-reassign */

  return array1.length === array2.length && array1.every(function (v, i) {
    return compareFunction(v, array2[i]);
  });
}

/**
 * Returns a random 32-bit integer (signed or unsigned) in the given range.
 * @memberof Utils
 * @param {number} min Minimum value (included).
 * @param {number} max Maximum value (excluded).
 * @returns {number}
 */
function getRandomInt(min, max) {
  var range = max - min;
  var bitLength = 32 - (0, _clz2.default)(range);

  // Avoid number re-generation by adding an extra byte
  var byteLength = (bitLength >> 3) + 1;
  var randomRange = Math.pow(256, byteLength);

  // Get the lowest value which is not part of the equal distribution range
  var firstTooHighValue = randomRange - randomRange % range;

  var result = void 0;
  do {
    // Generate a random number with the given byte length
    result = _crypto2.default.randomBytes(byteLength).readUIntBE(0, byteLength);
  } while (
  // Ensure equal distribution
  result >= firstTooHighValue);

  // Place the result in the initial range and offset it by the minimum value
  return result % range + min;
}

/**
 * Returns a random BigInt in the given range.
 * @memberof Utils
 * @param {BigInt} min Minimum value (included).
 * @param {BigInt} max Maximum value (excluded).
 * @returns {BigInt}
 */
function getRandomBigInt(min, max) {
  var range = max.redSub(min);

  // Avoid number re-generation by adding an extra byte
  var byteLength = range.byteLength() + 1;
  var randomRange = new _bn2.default(256).ishln(range.byteLength() << 3);

  // Get the lowest value which is not part of the equal distribution range
  var firstTooHighValue = randomRange.sub(randomRange.mod(range));

  var result = void 0;
  do {
    // Generate a random number with the given byte length
    result = new _bn2.default(_crypto2.default.randomBytes(byteLength));
  } while (
  // Ensure equal distribution
  result.gte(firstTooHighValue));

  // Place the result in the initial range and offset it by the minimum value
  return result.mod(range).toRed(_config2.default.ec.curve.red).redIAdd(min);
}

/**
 * Returns an array of random secrets.
 * @memberof Utils
 * @param {number} [amount] Amount of secrets to be generated.
 * @returns {BigInt[]}
 */
function getRandomSecrets() {
  var amount = arguments.length <= 0 || arguments[0] === undefined ? _config2.default.cardsInDeck + 1 : arguments[0];

  return (0, _from2.default)(new Array(amount), function () {
    return getRandomBigInt(_config2.default.ec.curve.one, _config2.default.ecRedN);
  });
}

/**
 * Returns an array of random points.
 * @memberof Utils
 * @param {number} amount Amount of points to be generated.
 * @returns {Point[]}
 */
function getRandomPoints() {
  var amount = arguments.length <= 0 || arguments[0] === undefined ? _config2.default.cardsInDeck : arguments[0];

  return getRandomSecrets(amount).map(function (secret) {
    return _config2.default.ec.g.mul(secret.fromRed());
  });
}

/**
 * Returns the hash of a given secret.
 * @memberof Utils
 * @param {BigInt} secret Secret to calculate the hash of.
 * @param {string} [algorithm] Hashing algorithm to be used.
 * @returns {string}
 */
function getSecretHash(secret) {
  var algorithm = arguments.length <= 1 || arguments[1] === undefined ? _config2.default.hashAlgorithm : arguments[1];

  return _crypto2.default.createHash(algorithm).update(secret.toString(16, 2)).digest('hex');
}

/**
 * Shuffles the elements of an array.
 * @memberof Utils
 * @param {T[]} array Array to be shuffled.
 * @returns {T[]}
 */
function shuffleArray(array) {
  var result = [].concat((0, _toConsumableArray3.default)(array));

  // Perform Durstenfeld shuffle
  for (var i = array.length - 1; i > 0; --i) {
    // Generate a random integer in [0, i] deterministically
    var j = getRandomInt(0, i);

    // Swap result[i] and result[j]
    var _ref = [result[j], result[i]];
    result[i] = _ref[0];
    result[j] = _ref[1];
  }

  return result;
}

/**
 * Converts a point into a JSON-serializable object.
 * @memberof Utils
 * @param {Point} point Point to be converted.
 * @returns {PointJSON}
 */
function pointToJSON(point) {
  return {
    x: point.x.toString(16, 2),
    y: point.y.toString(16, 2)
  };
}

/**
 * Sorts the given points in ascending order.
 * @memberof Utils
 * @param {Point[]} points Points to be sorted.
 * @returns {Point[]}
 */
function sortPoints(points) {
  return points.sort(function (p1, p2) {
    var xCmp = p1.x.cmp(p2.x);
    return xCmp === 0 ? p1.y.cmp(p2.y) : xCmp;
  });
}
//# sourceMappingURL=utils.js.map